import logging
from typing import Literal

import httpx
from httpx import Response, Timeout
from pydantic import BaseModel, Field
from unique_toolkit.agentic.tools.config import get_configuration_dict

from unique_web_search.client_settings import get_jina_search_settings
from unique_web_search.services.search_engine import (
    BaseSearchEngineConfig,
    SearchEngine,
    SearchEngineType,
)
from unique_web_search.services.search_engine.schema import (
    WebSearchResult,
)

logger = logging.getLogger(f"PythonAssistantCoreBundle.{__name__}")


class JinaSearchOptionalParams(BaseModel):
    model_config = get_configuration_dict()

    # Request body parameters
    gl: str | None = Field(
        default=None,
        description="The country to use for the search. It's a two-letter country code.",
    )
    location: str | None = Field(
        default=None,
        description="From where you want the search query to originate. It is recommended to specify location at the city level.",
    )
    hl: str | None = Field(
        default=None,
        description="The language to use for the search. It's a two-letter language code.",
    )
    num: int | None = Field(
        default=None,
        description="Sets maximum results returned. Using num may cause latency and exclude specialized result types.",
    )
    page: int | None = Field(
        default=None,
        description="The result offset. It skips the given number of results. It's used for pagination.",
    )

    # Header parameters (X-* headers)
    x_site: str | None = Field(
        default=None,
        alias="X-Site",
        description="For in-site searches limited to the given domain",
    )
    x_with_links_summary: str | None = Field(
        default=None,
        alias="X-With-Links-Summary",
        description="'all' to gather all links or 'true' to gather unique links at the end of the response",
    )
    x_with_images_summary: str | None = Field(
        default=None,
        alias="X-With-Images-Summary",
        description="'all' to gather all images or 'true' to gather unique images at the end of the response",
    )
    x_retain_images: str | None = Field(
        default=None,
        alias="X-Retain-Images",
        description="Use 'none' to remove all images from the response",
    )
    x_no_cache: bool | None = Field(
        default=None,
        alias="X-No-Cache",
        description="Set to true to bypass cache and retrieve real-time data",
    )
    x_with_generated_alt: bool | None = Field(
        default=None,
        alias="X-With-Generated-Alt",
        description="Set to true to generate captions for images without alt tags",
    )
    x_respond_with: str | None = Field(
        default=None,
        alias="X-Respond-With",
        description="Use 'no-content' to exclude page content from the response",
    )
    x_with_favicon: bool | None = Field(
        default=None,
        alias="X-With-Favicon",
        description="Set to true to include favicon of the website in the response",
    )
    x_return_format: str | None = Field(
        default=None,
        alias="X-Return-Format",
        description="'markdown', 'html', 'text', 'screenshot', or 'pageshot' for URL of full-page screenshot",
    )
    x_engine: str | None = Field(
        default=None,
        alias="X-Engine",
        description="Specifies the engine to retrieve/parse content. Use 'browser' for best quality or 'direct' for speed",
    )
    x_with_favicons: bool | None = Field(
        default=None,
        alias="X-With-Favicons",
        description="Set to true to fetch favicon of each URL in SERP and include them in response",
    )
    x_timeout: int | None = Field(
        default=None,
        alias="X-Timeout",
        description="Specifies the maximum time (in seconds) to wait for the webpage to load",
    )
    x_set_cookie: str | None = Field(
        default=None,
        alias="X-Set-Cookie",
        description="Forwards custom cookie settings when accessing the URL, useful for pages requiring authentication",
    )
    x_proxy_url: str | None = Field(
        default=None,
        alias="X-Proxy-Url",
        description="Utilizes your proxy to access URLs, helpful for pages accessible only through specific proxies",
    )
    x_locale: str | None = Field(
        default=None,
        alias="X-Locale",
        description="Controls the browser locale to render the page. Websites may serve different content based on locale",
    )


class JinaConfig(BaseSearchEngineConfig[SearchEngineType.JINA]):
    search_engine_name: Literal[SearchEngineType.JINA] = SearchEngineType.JINA

    custom_search_config: JinaSearchOptionalParams = JinaSearchOptionalParams()


class JinaSearchParams(BaseModel):
    q: str = Field(description="Search query")
    gl: str | None = Field(
        default=None,
        description="The country to use for the search. It's a two-letter country code.",
    )
    location: str | None = Field(
        default=None,
        description="From where you want the search query to originate.",
    )
    hl: str | None = Field(
        default=None,
        description="The language to use for the search. It's a two-letter language code.",
    )
    num: int | None = Field(
        default=None,
        description="Sets maximum results returned.",
    )
    page: int | None = Field(
        default=None,
        description="The result offset for pagination.",
    )


class JinaSearch(SearchEngine[JinaConfig]):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.is_configured = get_jina_search_settings().is_configured

    @property
    def requires_scraping(self) -> bool:
        return False

    def _get_request_params(self, query, **kwargs) -> dict:
        """Get the request parameters for Jina Search API."""
        # Ensure required settings are available
        jina_search_settings = get_jina_search_settings()
        assert jina_search_settings.api_key is not None

        # Separate body parameters from header parameters
        config_dict = self.config.custom_search_config.model_dump(
            mode="json", exclude_none=True, by_alias=True
        )

        # Body parameters (non-header fields)
        body_params = {
            k: v
            for k, v in config_dict.items()
            if v is not None and not k.startswith("X-")
        }

        # Header parameters (X-* fields)
        header_params = {
            k: str(v).lower() if v is not None and isinstance(v, bool) else v
            for k, v in config_dict.items()
            if k.startswith("X-")
        }

        # Create search parameters for the request body
        search_params = JinaSearchParams(
            q=query,
            num=self.config.fetch_size,
            **body_params,
        )

        # Base headers
        headers = {
            "Authorization": f"Bearer {jina_search_settings.api_key}",
            "Content-Type": "application/json",
            "Accept": "application/json",
        }

        # Add optional headers
        headers.update(header_params)

        params = {
            "url": jina_search_settings.search_api_endpoint,
            "headers": headers,
            "json": search_params.model_dump(exclude_none=True),
        }

        return params

    def _extract_urls(self, response: Response) -> list[WebSearchResult]:
        """Extract URLs from Jina Search API response."""
        try:
            results = response.json()

            # Handle error responses
            if response.status_code != 200:
                logger.error(f"Jina API error: {results}")
                return []

            # Extract data from Jina API response
            data = results.get("data", [])
            if not data:
                logger.warning("No search results found in Jina API response")
                return []

            extracted_results = []
            for item in data:
                try:
                    search_result = WebSearchResult(
                        url=item["url"],
                        title=item.get("title", ""),
                        snippet=item.get("description", ""),
                        content=item.get("content", ""),
                    )
                    extracted_results.append(search_result)
                except KeyError as e:
                    logger.warning(f"Missing required field in search result: {e}")
                    continue
                except Exception as e:
                    logger.warning(f"Error processing search result: {e}")
                    continue

            return extracted_results

        except Exception as e:
            logger.error(f"Failed to parse Jina API response: {e}")
            return []

    async def _perform_web_search_request(self, query: str, **kwargs) -> Response:
        """Send a request to the Jina Search API.

        Args:
            query: The search query.

        Returns:
            Response: The HTTP response from Jina API.
        """
        request_params = self._get_request_params(query=query, **kwargs)

        # Use POST method as required by Jina Search API
        async with httpx.AsyncClient() as client:
            response = await client.post(**request_params, timeout=Timeout(timeout=120))
        return response

    # TODO: Find a tracking solution
    # @track(
    #     tags=["jina_search"],
    # )
    async def search(self, query: str, **kwargs) -> list[WebSearchResult]:
        """Search the web for the given query using Jina Search API."""
        try:
            response = await self._perform_web_search_request(query=query, **kwargs)
            search_results = self._extract_urls(response=response)
            logger.info(f"Found {len(search_results)} URLs")

        except Exception as e:
            logger.exception(f"Failed to extract URLs from Jina search response: {e}")
            search_results = []

        return search_results



from enum import StrEnum
from typing import Literal

from pydantic import BaseModel, ConfigDict, Field, create_model


class WebSearchToolParameters(BaseModel):
    """Parameters for the Websearch tool."""

    model_config = ConfigDict(extra="forbid")
    query: str
    date_restrict: str | None

    @classmethod
    def from_tool_parameter_query_description(
        cls, query_description: str, date_restrict_description: str | None
    ) -> type["WebSearchToolParameters"]:
        """Create a new model with the query field."""
        return create_model(
            cls.__name__,
            query=(str, Field(description=query_description)),
            date_restrict=(
                str | None,
                Field(description=date_restrict_description),
            ),
            __base__=cls,
        )


class StepType(StrEnum):
    SEARCH = "search"
    READ_URL = "read_url"


class Step(BaseModel):
    step_type: Literal[StepType.SEARCH, StepType.READ_URL]
    objective: str = Field(description="The objective of the step")
    query_or_url: str = Field(
        description="The input for this step: either an optimized search query (for search steps) or a URL to read (for read_url steps)."
    )


class WebSearchPlan(BaseModel):
    model_config = ConfigDict(extra="forbid")
    objective: str = Field(description="The objective of the plan")
    query_analysis: str = Field(
        description="Analysis of the user's query and what information is needed"
    )
    steps: list[Step] = Field(description="Steps to execute")
    expected_outcome: str = Field(description="Expected outcome")



import logging
import os
import sys
from pathlib import Path

from pydantic_settings import BaseSettings, SettingsConfigDict

# TODO: Clean this up from monorepo. Its mixing everything

logger = logging.getLogger(f"PythonAssistantCoreBundle.{__name__}")


class Base(BaseSettings):
    env: str | None = None
    log_level: str | None = None
    tiktoken_cache_dir: str = "./tiktoken_cache/"

    # Active search engines
    active_search_engines: list[str] = ["google"]

    # Bing settings
    bing_search_v7_subscription_key: str | None = None
    bing_api_endpoint: str | None = None

    # Google Search settings
    google_search_api_key: str | None = None
    google_search_api_endpoint: str | None = None
    google_search_engine_id: str | None = None

    # Jina API settings
    jina_api_key: str | None = None
    jina_search_api_endpoint: str = "https://s.jina.ai/"
    jina_reader_api_endpoint: str = "https://r.jina.ai/"

    # Firecrawl API settings
    firecrawl_api_key: str | None = None

    # Tavily API settings
    tavily_api_key: str | None = None

    # Brave Search API settings
    brave_search_api_key: str | None = None
    brave_search_api_endpoint: str | None = None


class Settings(Base):
    model_config = SettingsConfigDict(
        env_file=Path(os.getcwd()) / ".env", extra="ignore"
    )


class TestSettings(Base):
    model_config = SettingsConfigDict(
        env_file=Path(os.getcwd()) / "tests/test.env", extra="ignore"
    )


def get_settings() -> Base:
    """
    Dynamically load settings, switching to test environment if running under pytest.

    :return: Settings instance.
    """
    if "pytest" in sys.modules:
        # Dynamically adjust to use `test.env` if running under pytest
        return TestSettings()
    return Settings()


env_settings: Base = get_settings()



from logging import getLogger

from pydantic import BaseModel, Field, model_validator
from pydantic.json_schema import SkipJsonSchema
from unique_toolkit._common.chunk_relevancy_sorter.config import (
    ChunkRelevancySortConfig,
)
from unique_toolkit._common.default_language_model import DEFAULT_GPT_4o
from unique_toolkit._common.feature_flags.schema import (
    FeatureExtendedSourceSerialization,
)
from unique_toolkit._common.validators import LMI, get_LMI_default_field
from unique_toolkit.agentic.evaluation.schemas import EvaluationMetricName
from unique_toolkit.agentic.tools.config import get_configuration_dict
from unique_toolkit.agentic.tools.schemas import BaseToolConfig
from unique_toolkit.language_model.infos import ModelCapabilities

from unique_web_search.prompts import (
    DEFAULT_TOOL_FORMAT_INFORMATION_FOR_SYSTEM_PROMPT,
    REFINE_QUERY_SYSTEM_PROMPT,
    RESTRICT_DATE_DESCRIPTION,
    TOOL_DESCRIPTIONS,
    TOOL_DESCRIPTIONS_FOR_SYSTEM_PROMPT,
)
from unique_web_search.services.content_processing.config import (
    ContentProcessorConfig,
)
from unique_web_search.services.crawlers import (
    BasicCrawlerConfig,
    CrawlerConfigTypes,
)
from unique_web_search.services.executors.web_search_v1_executor import RefineQueryMode
from unique_web_search.services.search_engine import (
    get_config_types_from_names,
    get_default_search_engine_config,
)
from unique_web_search.settings import env_settings

logger = getLogger(__name__)

DEFAULT_MODEL_NAME = DEFAULT_GPT_4o

ActivatedSearchEngine = get_config_types_from_names(env_settings.active_search_engines)
DefaultSearchEngine = get_default_search_engine_config(
    env_settings.active_search_engines
)


class AnswerGenerationConfig(BaseModel):
    model_config = get_configuration_dict()

    limit_token_sources: int = Field(
        default=10000,
        description="Token Source Limit",
    )
    max_chunks_to_consider: int = Field(
        default=20,
        description="Token Source Limit",
    )
    number_history_interactions_included: int = Field(
        default=2,
        description="Number of history interactions included",
    )


class WebSearchV2(BaseModel):
    model_config = get_configuration_dict()

    enabled: bool = Field(
        default=False,
        description="Enable or disable the WebSearch V2 tool. Set to True to activate the step-based web search executor.",
    )
    max_steps: int = Field(
        default=5,
        description="Maximum number of sequential steps (searches or URL reads) allowed in a single WebSearch V2 plan.",
    )
    tool_description: str = Field(
        default=TOOL_DESCRIPTIONS["v2"],
        description="Information to help the language model decide when to select this tool; describes the tool's general purpose and when it is relevant.",
    )
    tool_description_for_system_prompt: str = Field(
        default=TOOL_DESCRIPTIONS_FOR_SYSTEM_PROMPT["v2"],
        description="Description of the tool's capabilities, intended for inclusion in system prompts to inform the language model what the tool can do.",
    )
    tool_format_information_for_system_prompt: str = Field(
        default=DEFAULT_TOOL_FORMAT_INFORMATION_FOR_SYSTEM_PROMPT,
        description="Instructions for the language model on how to reference and organize information from the tool in its response.",
    )


class WebSearchV1(BaseModel):
    model_config = get_configuration_dict()

    refine_query_mode: RefineQueryMode = Field(
        default=RefineQueryMode.BASIC,
        description="Query refinement strategy for WebSearch V1. Determines how user queries are improved before searching (e.g., BASIC, ADVANCED).",
    )
    max_queries: int = Field(
        default=5,
        description="Maximum number of search queries that WebSearch V1 will issue per user request.",
    )


class ExperimentalFeatures(FeatureExtendedSourceSerialization):
    v1_mode: WebSearchV1 = Field(
        default_factory=WebSearchV1,
        description=(
            "Configuration options for WebSearch V1 mode. "
            "Controls the behavior and parameters of the original web search tool, "
            "including query refinement and search limits."
        ),
    )

    v2_mode: WebSearchV2 = Field(
        default_factory=WebSearchV2,
        description=(
            "Configuration options for WebSearch V2 mode. "
            "Enables and customizes the new step-based web search executor, "
            "allowing for advanced planning and multi-step research workflows."
        ),
    )


class QueryRefinementConfig(BaseModel):
    model_config = get_configuration_dict()

    enabled: bool = Field(
        default=True,
        description="Whether to enable the refined query",
    )
    system_prompt: str = Field(
        default=REFINE_QUERY_SYSTEM_PROMPT,
        description="The system prompt to refine the query",
    )


class WebSearchToolParametersConfig(BaseModel):
    model_config = get_configuration_dict()

    query_description: str = Field(
        default="The search query to issue to the web.",
        description="The tool parameter query description",
    )
    date_restrict_description: str = Field(
        default=RESTRICT_DATE_DESCRIPTION,
        description="The tool parameter date restrict description",
    )


class WebSearchConfig(BaseToolConfig):
    language_model: LMI = get_LMI_default_field(DEFAULT_MODEL_NAME)

    limit_token_sources: SkipJsonSchema[int] = Field(
        default=60_000,  # TODO: Remove SkipJsonSchema once UI (Spaces 2.0) can be configured to not include certain fields
        description="Token Source Limit",
    )
    percentage_of_input_tokens_for_sources: float = Field(
        default=0.4,
        description="The percentage of the maximum input tokens of the language model to use for the tool response.",
        ge=0.0,
        le=1.0,
    )
    language_model_max_input_tokens: SkipJsonSchema[int | None] = Field(
        default=None,
        description="Language model maximum input tokens",
    )

    search_engine_config: ActivatedSearchEngine = Field(  # type: ignore
        default_factory=DefaultSearchEngine,  # type: ignore
        description="Search Engine Configuration",
        discriminator="search_engine_name",
        title="Search Engine Configuration",
    )

    crawler_config: CrawlerConfigTypes = Field(
        default_factory=BasicCrawlerConfig,
        title="Crawler Configuration",
        description="Crawler configuration.",
        discriminator="crawler_type",
    )

    content_processor_config: ContentProcessorConfig = Field(
        default_factory=ContentProcessorConfig,
        description="Content processor configuration",
        title="Content Processor Configuration",
    )

    chunk_relevancy_sort_config: ChunkRelevancySortConfig = Field(
        default_factory=ChunkRelevancySortConfig,
        description="Chunk Relevancy Sort Configuration",
        title="Chunk Relevancy Sort Configuration",
    )

    evaluation_check_list: list[EvaluationMetricName] = Field(
        default=[
            EvaluationMetricName.HALLUCINATION,
        ],
        description="Check list of evaluations executed conditionally after the answer is generated",
        title="Evaluation Check List",
    )

    query_refinement_config: QueryRefinementConfig = Field(
        default_factory=QueryRefinementConfig,
        description="The query refinement configuration",
        title="Query Refinement Configuration",
    )

    tool_parameters_config: WebSearchToolParametersConfig = Field(
        default_factory=WebSearchToolParametersConfig,
        description="Tool parameters configuration",
        title="Tool Parameters Configuration",
    )

    tool_description: str = Field(
        default=TOOL_DESCRIPTIONS["v1"],
        description="Tool description",
    )

    tool_description_for_system_prompt: str = Field(
        default=TOOL_DESCRIPTIONS_FOR_SYSTEM_PROMPT["v1"],
    )

    tool_format_information_for_system_prompt: str = Field(
        default=DEFAULT_TOOL_FORMAT_INFORMATION_FOR_SYSTEM_PROMPT,
    )

    experimental_features: ExperimentalFeatures = ExperimentalFeatures()

    debug: bool = Field(
        default=False,
        description="Whether to enable the debug mode",
    )

    @model_validator(mode="after")
    def disable_query_refinement_if_no_structured_output(self):
        if ModelCapabilities.STRUCTURED_OUTPUT not in self.language_model.capabilities:
            self.query_refinement_config.enabled = False
            logger.warning(
                "The language model does not support structured output. Query refinement is disabled."
            )
        return self



from unique_toolkit._common.pydantic_helpers import (
    model_title_generator,
)


def clean_model_title_generator(model: type) -> str:
    title = model_title_generator(model)
    return title.replace("Config", "").strip()


def experimental_model_title_generator(model: type) -> str:
    title = clean_model_title_generator(model)
    return title.replace("Config", "").strip() + " (Experimental)"



from urllib.parse import urlparse

from pydantic import BaseModel, Field


def extract_root_domain(url: str) -> str:
    parsed_url = urlparse(url)
    domain = parsed_url.netloc
    domain = domain.replace("www.", "")
    return domain


class WebSearchResult(BaseModel):
    url: str
    title: str
    snippet: str = Field(
        ...,
        description="A short description of the content found on this website",
    )
    content: str = Field(
        default="",
        description="The content of the website",
    )

    @property
    def display_link(self):
        return extract_root_domain(self.url)



import logging
from enum import StrEnum
from time import time
from typing import Callable, Literal, Optional, overload, override

from pydantic import Field
from unique_toolkit import LanguageModelService
from unique_toolkit._common.chunk_relevancy_sorter.config import (
    ChunkRelevancySortConfig,
)
from unique_toolkit._common.chunk_relevancy_sorter.service import ChunkRelevancySorter
from unique_toolkit._common.utils.structured_output.schema import StructuredOutputModel
from unique_toolkit._common.validators import LMI
from unique_toolkit.agentic.tools.tool_progress_reporter import (
    ToolProgressReporter,
)
from unique_toolkit.content import ContentChunk
from unique_toolkit.language_model import LanguageModelFunction
from unique_toolkit.language_model.builder import MessagesBuilder

from unique_web_search.schema import WebSearchToolParameters
from unique_web_search.services.content_processing import ContentProcessor, WebPageChunk
from unique_web_search.services.crawlers import CrawlerTypes
from unique_web_search.services.executors.base_executor import BaseWebSearchExecutor
from unique_web_search.services.search_engine import SearchEngineTypes
from unique_web_search.services.search_engine.schema import (
    WebSearchResult,
)
from unique_web_search.utils import (
    StepDebugInfo,
    WebSearchDebugInfo,
    query_params_to_human_string,
)

logger = logging.getLogger(f"PythonAssistantCoreBundle.{__name__}")


class RefineQueryMode(StrEnum):
    ADVANCED = "advanced"
    BASIC = "basic"


class RefinedQuery(StructuredOutputModel):
    """A refined query."""

    objective: str = Field(description="The objective of the original query")
    refined_query: str = Field(
        description="The refined query optimized for the search engine."
    )


class RefinedQueries(StructuredOutputModel):
    """A refined query."""

    objective: str = Field(description="The overall objective of the original query")
    refined_queries: list[RefinedQuery] = Field(
        description="The refined queries optimized for the search engine."
    )


@overload
async def query_generation_agent(
    query: str,
    language_model_service: LanguageModelService,
    language_model: LMI,
    system_prompt: str,
    mode: Literal[RefineQueryMode.BASIC],
) -> RefinedQueries: ...


@overload
async def query_generation_agent(
    query: str,
    language_model_service: LanguageModelService,
    language_model: LMI,
    system_prompt: str,
    mode: Literal[RefineQueryMode.ADVANCED],
) -> RefinedQuery: ...


async def query_generation_agent(
    query: str,
    language_model_service: LanguageModelService,
    language_model: LMI,
    system_prompt: str,
    mode: RefineQueryMode,
) -> RefinedQuery | RefinedQueries:
    """Refine the query to be more specific and relevant to the user's question."""
    messages = (
        MessagesBuilder()
        .system_message_append(system_prompt)
        .user_message_append(query)
        .build()
    )

    if mode == RefineQueryMode.BASIC:
        structured_output_model = RefinedQueries
    else:
        structured_output_model = RefinedQuery

    response = await language_model_service.complete_async(
        messages,
        model_name=language_model.name,
        structured_output_model=structured_output_model,
        structured_output_enforce_schema=True,
    )

    parsed_response = response.choices[0].message.parsed
    if parsed_response is None:
        raise ValueError("Failed to parse insights from LLM response")

    return structured_output_model.model_validate(parsed_response)


class WebSearchV1Executor(BaseWebSearchExecutor):
    """Executes research plans step by step."""

    @override
    def __init__(
        self,
        company_id: str,
        language_model_service: LanguageModelService,
        language_model: LMI,
        search_service: SearchEngineTypes,
        crawler_service: CrawlerTypes,
        content_processor: ContentProcessor,
        chunk_relevancy_sorter: ChunkRelevancySorter | None,
        chunk_relevancy_sort_config: ChunkRelevancySortConfig,
        content_reducer: Callable[[list[WebPageChunk]], list[WebPageChunk]],
        tool_call: LanguageModelFunction,
        tool_parameters: WebSearchToolParameters,
        refine_query_system_prompt: str,
        debug_info: WebSearchDebugInfo,
        tool_progress_reporter: Optional[ToolProgressReporter] = None,
        mode: RefineQueryMode = RefineQueryMode.BASIC,
        max_queries: int = 10,
    ):
        super().__init__(
            search_service=search_service,
            language_model_service=language_model_service,
            language_model=language_model,
            crawler_service=crawler_service,
            tool_call=tool_call,
            tool_parameters=tool_parameters,
            company_id=company_id,
            content_processor=content_processor,
            chunk_relevancy_sorter=chunk_relevancy_sorter,
            chunk_relevancy_sort_config=chunk_relevancy_sort_config,
            debug_info=debug_info,
            content_reducer=content_reducer,
            tool_progress_reporter=tool_progress_reporter,
        )
        self.mode = mode
        self.tool_parameters = tool_parameters
        self.refine_query_system_prompt = refine_query_system_prompt
        self.max_queries = max_queries

    async def run(self) -> list[ContentChunk]:
        query = self.tool_parameters.query
        date_restrict = self.tool_parameters.date_restrict

        self.notify_name = "**Refining Query**"
        self.notify_message = query_params_to_human_string(query, date_restrict)
        await self.notify_callback()
        refined_queries, objective = await self._refine_query(query)

        web_search_results = []
        for index, refined_query in enumerate(refined_queries):
            if len(refined_queries) > 1:
                self.notify_name = (
                    f"**Searching Web {index + 1}/{len(refined_queries)}**"
                )
            else:
                self.notify_name = "**Searching Web**"

            self.notify_message = query_params_to_human_string(
                refined_query, date_restrict
            )
            await self.notify_callback()

            search_results = await self._search(
                refined_query, date_restrict=date_restrict
            )
            web_search_results.extend(search_results)

        if self.search_service.requires_scraping:
            self.notify_name = "**Crawling URLs**"
            self.notify_message = f"{len(web_search_results)} URLs to fetch"
            await self.notify_callback()
            crawl_results = await self._crawl(web_search_results)
            for web_search_result, crawl_result in zip(
                web_search_results, crawl_results
            ):
                web_search_result.content = crawl_result

        self.notify_name = "**Analyzing Web Pages**"
        self.notify_message = objective
        await self.notify_callback()

        content_results = await self._content_processing(objective, web_search_results)

        if self.chunk_relevancy_sorter:
            self.notify_name = "**Resorting Sources**"
            self.notify_message = objective
            await self.notify_callback()

        relevant_sources = await self._select_relevant_sources(
            objective, content_results
        )

        return relevant_sources

    async def _refine_query(self, query: str) -> tuple[list[str], str]:
        start_time = time()
        refined_query = await query_generation_agent(
            query,
            self.language_model_service,
            self.language_model,
            self.refine_query_system_prompt,
            self.mode,
        )
        end_time = time()
        delta_time = end_time - start_time

        if isinstance(refined_query, RefinedQuery):
            queries = [refined_query.refined_query]
        elif isinstance(refined_query, RefinedQueries):
            queries = [
                refined_query.refined_query
                for refined_query in refined_query.refined_queries
            ]
        else:
            raise ValueError("Invalid refined query")
        self.debug_info.steps.append(
            StepDebugInfo(
                step_name="refine_query",
                execution_time=delta_time,
                config=self.mode.name,
                extra={
                    "query": query,
                    "refined_queries": queries,
                },
            )
        )

        return queries, refined_query.objective

    async def _search(
        self, query: str, date_restrict: str | None
    ) -> list[WebSearchResult]:
        start_time = time()
        logger.info(f"Company {self.company_id} Searching with {self.search_service}")
        search_results = await self.search_service.search(
            query, date_restrict=date_restrict
        )
        end_time = time()
        delta_time = end_time - start_time
        logger.info(
            f"Searched with {self.search_service} completed in {delta_time} seconds"
        )
        self.debug_info.steps.append(
            StepDebugInfo(
                step_name="search",
                execution_time=delta_time,
                config=self.search_service.config.search_engine_name.name,
                extra={
                    "query": query,
                    "date_restrict": date_restrict,
                    "number_of_results": len(search_results),
                    "urls": [result.url for result in search_results],
                },
            )
        )
        return search_results

    async def _crawl(self, web_search_results: list[WebSearchResult]) -> list[str]:
        start_time = time()
        logger.info(f"Company {self.company_id} Crawling with {self.crawler_service}")
        crawl_results = await self.crawler_service.crawl(
            [result.url for result in web_search_results]
        )
        end_time = time()
        delta_time = end_time - start_time
        logger.info(
            f"Crawled {len(web_search_results)} pages with {self.crawler_service} completed in {delta_time} seconds"
        )
        self.debug_info.steps.append(
            StepDebugInfo(
                step_name="crawl",
                execution_time=delta_time,
                config=self.crawler_service.config.crawler_type.name,
                extra={
                    "number_of_results": len(web_search_results),
                    "contents": [result.model_dump() for result in web_search_results],
                },
            )
        )
        return crawl_results

    def _enforce_max_queries(self, list_of_queries: list[str]) -> list[str]:
        if len(list_of_queries) > self.max_queries:
            logger.info(
                f"Company {self.company_id} Reducing number of queries to {self.max_queries}"
            )
            list_of_queries = list_of_queries[: self.max_queries]
        return list_of_queries



import logging
from typing import Literal

from httpx import AsyncClient, Response
from pydantic import BaseModel

from unique_web_search.client_settings import get_brave_search_settings
from unique_web_search.services.search_engine import (
    BaseSearchEngineConfig,
    SearchEngine,
    SearchEngineType,
)
from unique_web_search.services.search_engine.schema import (
    WebSearchResult,
)

logger = logging.getLogger(__name__)

HEADERS = {
    "Accept": "application/json",
    "Accept-Encoding": "gzip",
}

PAGINATION_SIZE = (
    20  # Brave Search API only supports a maximum of 20 results per request
)


def get_headers(api_key: str) -> dict[str, str]:
    return HEADERS.copy() | {"X-Subscription-Key": api_key}


class BraveSearchParameters(BaseModel):
    q: str
    count: int
    offset: int
    safesearch: Literal["strict", "moderate", "off"] = "strict"


class BraveSearchConfig(BaseSearchEngineConfig[SearchEngineType.BRAVE]):
    search_engine_name: Literal[SearchEngineType.BRAVE] = SearchEngineType.BRAVE


class BraveSearch(SearchEngine[BraveSearchConfig]):
    def __init__(
        self,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.is_configured = get_brave_search_settings().is_configured

    @property
    def requires_scraping(self) -> bool:
        return True

    async def search(self, query: str, **kwargs) -> list[WebSearchResult]:
        search_results = []
        fetch_size = self.config.fetch_size

        for start_index in range(0, fetch_size, PAGINATION_SIZE):
            effective_num_fetch = min(fetch_size - start_index, PAGINATION_SIZE)
            params = BraveSearchParameters(
                q=query, count=effective_num_fetch, offset=start_index
            )

            response = await self._perform_web_search_request(params=params)
            search_results.extend(self._extract_urls(response.json()))

        return search_results

    async def _perform_web_search_request(
        self, params: BraveSearchParameters
    ) -> Response:
        """Send a request to the search engine.

        Args:
            query: The query.
            start_index: The start index.

        Returns:
            list[dict]: The search results.

        """
        api_endpoint = get_brave_search_settings().api_endpoint
        api_key = get_brave_search_settings().api_key
        assert api_key is not None and api_endpoint is not None

        async with AsyncClient() as client:
            response = await client.get(
                api_endpoint,
                params=params.model_dump(exclude_none=True),
                headers=get_headers(api_key),
            )
        return response

    def _extract_urls(self, brave_response: dict) -> list[WebSearchResult]:
        search_results = []
        if "web" in brave_response and brave_response["web"] is not None:
            search_results.extend(brave_response["web"]["results"])
        if "news" in brave_response and brave_response["news"] is not None:
            search_results.extend(brave_response["news"]["results"])
        if not search_results:
            logger.warning("No search results found in Brave search response")
            return []

        return [
            WebSearchResult(
                url=item["url"],
                title=item["title"],
                snippet=self._safe_load_key(item, "description"),
            )
            for item in search_results
        ]

    @staticmethod
    def _safe_load_key(brave_response: dict, key: str, default: str = "") -> str:
        try:
            return brave_response[key]
        except KeyError:
            logger.warning(f"Key {key} not found in Brave search response")
            return default



import asyncio
from typing import Literal

from tavily import AsyncTavilyClient

from unique_web_search.client_settings import get_tavily_search_settings
from unique_web_search.services.crawlers.base import (
    BaseCrawler,
    BaseCrawlerConfigExperimental,
    CrawlerType,
)


class TavilyCrawlerConfig(BaseCrawlerConfigExperimental[CrawlerType.TAVILY]):
    crawler_type: Literal[CrawlerType.TAVILY] = CrawlerType.TAVILY
    depth: Literal["basic", "advanced"] = "advanced"


class TavilyCrawler(BaseCrawler[TavilyCrawlerConfig]):
    def __init__(self, config: TavilyCrawlerConfig):
        super().__init__(config)

    # TODO: find a tracking solution
    # @track(
    #     tags=["tavily", "scrape"],
    # )
    async def crawl(self, urls: list[str]) -> list[str]:
        api_key = get_tavily_search_settings().api_key
        assert api_key is not None

        client = AsyncTavilyClient(api_key=api_key)

        # Tavily only supports up to 20 URLs per request, so we need to batch the requests
        response = {"results": [], "failed_results": []}
        batch_size = 20

        async def process_batch(batch_urls: list[str]) -> dict:
            batch_response = await client.extract(
                urls=batch_urls,
                format="markdown",
                include_images=False,
                extract_depth=self.config.depth,
                timeout=self.config.timeout,
                include_favicon=False,
            )
            return batch_response

        tasks = [
            process_batch(urls[i : i + batch_size])
            for i in range(0, len(urls), batch_size)
        ]
        results = await asyncio.gather(*tasks)

        for result in results:
            response["results"].extend(result.get("results", []))
            response["failed_results"].extend(result.get("failed_results", []))

        for i in range(0, len(urls), batch_size):
            batch_urls = urls[i : i + batch_size]
            batch_response = await client.extract(
                urls=batch_urls,
                format="markdown",
                include_images=False,
                extract_depth=self.config.depth,
                timeout=self.config.timeout,
                include_favicon=False,
            )
            response["results"].extend(batch_response.get("results", []))
            response["failed_results"].extend(batch_response.get("failed_results", []))

        # Create a mapping from URL to content
        url_to_content = {}

        # Process successful results
        for result in response["results"]:
            url_to_content[result["url"]] = result["raw_content"]

        # Process failed results with error messages
        for failed_result in response["failed_results"]:
            url_to_content[failed_result["url"]] = f"Error: {failed_result['error']}"

        # Return results in the same order as input URLs, with fallback for missing URLs
        return [
            url_to_content.get(url, "Error: URL not found in response") for url in urls
        ]



from time import time

from typing_extensions import override
from unique_toolkit._common.chunk_relevancy_sorter.service import ChunkRelevancySorter
from unique_toolkit.agentic.evaluation.schemas import EvaluationMetricName
from unique_toolkit.agentic.history_manager.utils import transform_chunks_to_string
from unique_toolkit.agentic.tools.factory import ToolFactory
from unique_toolkit.agentic.tools.schemas import ToolCallResponse
from unique_toolkit.agentic.tools.tool import (
    AgentChunksHandler,
    LanguageModelMessage,
    Tool,
)
from unique_toolkit.agentic.tools.tool_progress_reporter import ProgressState
from unique_toolkit.language_model import LanguageModelToolMessage
from unique_toolkit.language_model.schemas import (
    LanguageModelFunction,
    LanguageModelToolDescription,
)

from unique_web_search.config import WebSearchConfig
from unique_web_search.schema import WebSearchPlan, WebSearchToolParameters
from unique_web_search.services.content_processing import ContentProcessor, WebPageChunk
from unique_web_search.services.crawlers import get_crawler_service
from unique_web_search.services.executors import (
    WebSearchV1Executor,
    WebSearchV2Executor,
)
from unique_web_search.services.search_engine import get_search_engine_service
from unique_web_search.utils import WebSearchDebugInfo, reduce_sources_to_token_limit


class WebSearchTool(Tool[WebSearchConfig]):
    name = "WebSearch"

    def __init__(self, configuration: WebSearchConfig, *args, **kwargs):
        super().__init__(configuration, *args, **kwargs)
        self.language_model = self.config.language_model

        self.search_engine_service = get_search_engine_service(
            self.config.search_engine_config,
            self.language_model_service,
            self.language_model,
        )
        self.crawler_service = get_crawler_service(self.config.crawler_config)
        self.chunk_relevancy_sorter = ChunkRelevancySorter(self.event)
        self.company_id = self.event.company_id
        self.chat_history_token_length = 0
        self.chat_history_chat_messages = self._chat_service.get_full_history()
        self.content_processor = ContentProcessor(
            event=self.event,
            config=self.config.content_processor_config,
            language_model=self.language_model,
        )
        self.debug = self.config.debug

        def content_reducer(web_page_chunks: list[WebPageChunk]) -> list[WebPageChunk]:
            return reduce_sources_to_token_limit(
                web_page_chunks,
                self.config.language_model_max_input_tokens,
                self.config.percentage_of_input_tokens_for_sources,
                self.config.limit_token_sources,
                self.language_model,
                self.chat_history_token_length,
            )

        self.content_reducer = content_reducer

    @override
    def tool_description(self) -> LanguageModelToolDescription:
        self.tool_parameter_calls = (
            WebSearchPlan
            if self.config.experimental_features.v2_mode.enabled
            else WebSearchToolParameters.from_tool_parameter_query_description(
                self.config.tool_parameters_config.query_description,
                self.config.tool_parameters_config.date_restrict_description,
            )
        )

        tool_description = (
            self.config.experimental_features.v2_mode.tool_description
            if self.config.experimental_features.v2_mode.enabled
            else self.config.tool_description
        )

        return LanguageModelToolDescription(
            name=self.name,
            description=tool_description,
            parameters=self.tool_parameter_calls,
        )

    def tool_description_for_system_prompt(self) -> str:
        if self.config.experimental_features.v2_mode.enabled:
            return self.config.experimental_features.v2_mode.tool_description_for_system_prompt.replace(
                "$max_steps", str(self.config.experimental_features.v2_mode.max_steps)
            )

        return self.config.tool_description_for_system_prompt

    def tool_format_information_for_system_prompt(self) -> str:
        return self.config.tool_format_information_for_system_prompt

    def evaluation_check_list(self) -> list[EvaluationMetricName]:
        return self.config.evaluation_check_list

    @override
    async def run(self, tool_call: LanguageModelFunction) -> ToolCallResponse:
        self.logger.info("Running the WebSearch tool")
        start_time = time()
        parameters = self.tool_parameter_calls.model_validate(
            tool_call.arguments,
        )

        debug_info = WebSearchDebugInfo(parameters=parameters.model_dump())
        executor = self._get_executor(tool_call, parameters, debug_info)

        try:
            content_chunks = await executor.run()
            debug_info.num_chunks_in_final_prompts = len(content_chunks)
            debug_info.execution_time = time() - start_time

            if self.tool_progress_reporter:
                await self.tool_progress_reporter.notify_from_tool_call(
                    tool_call=tool_call,
                    name=executor.notify_name,
                    message=executor.notify_message,
                    state=ProgressState.FINISHED,
                )

            return ToolCallResponse(
                id=tool_call.id,  # type: ignore
                name=self.name,
                debug_info=debug_info.model_dump(with_debug_details=self.debug),
                content_chunks=content_chunks,
            )
        except Exception as e:
            self.logger.exception(f"Error executing WebSearch tool: {e}")

            if self.tool_progress_reporter:
                await self.tool_progress_reporter.notify_from_tool_call(
                    tool_call=tool_call,
                    name=executor.notify_name,
                    message=executor.notify_message,
                    state=ProgressState.FAILED,
                )

            return ToolCallResponse(
                id=tool_call.id,  # type: ignore
                name=self.name,
                debug_info=debug_info.model_dump(with_debug_details=self.debug),
                error_message=str(e),
            )

    def _get_executor(
        self,
        tool_call: LanguageModelFunction,
        parameters: WebSearchPlan | WebSearchToolParameters,
        debug_info: WebSearchDebugInfo,
    ) -> WebSearchV1Executor | WebSearchV2Executor:
        if isinstance(parameters, WebSearchPlan):
            return WebSearchV2Executor(
                search_service=self.search_engine_service,
                language_model_service=self.language_model_service,
                language_model=self.language_model,
                crawler_service=self.crawler_service,
                tool_call=tool_call,
                tool_parameters=parameters,
                company_id=self.company_id,
                content_processor=self.content_processor,
                chunk_relevancy_sorter=self.chunk_relevancy_sorter,
                chunk_relevancy_sort_config=self.config.chunk_relevancy_sort_config,
                tool_progress_reporter=self.tool_progress_reporter,
                content_reducer=self.content_reducer,
                max_steps=self.config.experimental_features.v2_mode.max_steps,
                debug_info=debug_info,
            )
        elif isinstance(parameters, WebSearchToolParameters):
            return WebSearchV1Executor(
                search_service=self.search_engine_service,
                language_model_service=self.language_model_service,
                language_model=self.language_model,
                crawler_service=self.crawler_service,
                tool_call=tool_call,
                tool_parameters=parameters,
                company_id=self.company_id,
                mode=self.config.experimental_features.v1_mode.refine_query_mode,
                max_queries=self.config.experimental_features.v1_mode.max_queries,
                content_processor=self.content_processor,
                chunk_relevancy_sorter=self.chunk_relevancy_sorter,
                chunk_relevancy_sort_config=self.config.chunk_relevancy_sort_config,
                tool_progress_reporter=self.tool_progress_reporter,
                content_reducer=self.content_reducer,
                refine_query_system_prompt=self.config.query_refinement_config.system_prompt,
                debug_info=debug_info,
            )
        else:
            raise ValueError(f"Invalid parameters: {parameters}")

    def get_tool_call_result_for_loop_history(
        self,
        tool_response: ToolCallResponse,
        agent_chunks_handler: AgentChunksHandler,
    ) -> LanguageModelMessage:
        """Process the results of the tool.

        Args:
        ----
            tool_response: The tool response.
            loop_history: The loop history.

        Returns:
        -------
            The tool result to append to the loop history.

        """
        self.logger.debug(
            f"Appending tool call result to history: {tool_response.name}"
        )
        # Initialize content_chunks if None
        content_chunks = tool_response.content_chunks or []

        # Get the maximum source number in the loop history
        max_source_number = len(agent_chunks_handler.chunks)

        # Transform content chunks into sources to be appended to tool result
        sources, _ = transform_chunks_to_string(
            content_chunks,
            max_source_number,
            None,  # Use None for SourceFormatConfig
            self.config.experimental_features.full_sources_serialize_dump,
        )

        # Append the result to the history
        return LanguageModelToolMessage(
            content=sources,
            tool_call_id=tool_response.id,  # type: ignore
            name=tool_response.name,
        )

    def get_evaluation_checks_based_on_tool_response(
        self,
        tool_response: ToolCallResponse,
    ) -> list[EvaluationMetricName]:
        evaluation_check_list = self.evaluation_check_list()

        # Check if the tool response is empty
        if not tool_response.content_chunks:
            return []
        return evaluation_check_list


ToolFactory.register_tool(WebSearchTool, WebSearchConfig)
